#include <iostream>
#include <assert.h>
#include "scalar.h"
#include "vector.h"
#include "versor.h"
#include "point.h"
#include "quaternion.h"

// POD: Plain old data
struct Plane{
    Point p;
    Vector n;
};

struct Ray{
    Point start;
    Vector dir;
};

// TESTS
void unitTestSumIsCommutative() {
    Vector v(1, 2, 3), w(4, 5, 6);
    assert(isEqual(v+w, w+v));
}

void unitTestCross() {
    Vector v(1, 2, 3), w(4, 5, 6);
    assert(isEqual(cross(v, w), -cross(w, v)));
    // Resulting vector is orthogonal to both
    assert(isZero(dot(cross(v,w), v)));
    assert(isZero(dot(cross(v,w), w)));
}

void unitTestVersor() {
    Vector v(1, 2, 3);
    Versor d(v);
    assert(isEqual(d * length(v), v));
    assert(isEqual(length(d.asVector()), 1));
    assert(isEqual( angleBetween(Versor::up(), Versor::forward()), M_PI/2) );
}

void unitTestAngle() {
    Point a(3,6,3), b(8,1,4), c(0,7,2);
    assert(isEqual(
        length(b-a) / sin(angleBetween(a-c, b-c)),
        length(c-b) / sin(angleBetween(b-a, c-a))
    ));
}

// Reflect ray v off surface with normal n
Vector rayReflection(Vector v, Vector n) {
    return v - (2 * dot(n, v)) * n;
}


Point intersection(Plane plane, Ray ray) {
    Scalar k = dot(plane.p - ray.start, plane.n)/
               dot(ray.dir, plane.n);
    // todo: avoid division by 0
    return ray.start + k*ray.dir;
}

void unitTestQuaternionRotation2() {
    Quaternion q = Quaternion::axisAngle(
                Versor::up(),
                toRad(180)
                );
    assert(isEqual(q, Quaternion(0, 1, 0, 0)));

    Vector v(7, 2, 6);
    assert(isEqual(rotate(q, v), Vector(-7, 2, -6)));
 }

void unitTestQuaternionRotation() {
    Vector v0(15.6, -7.1 , 13);
    Quaternion q = Quaternion::axisAngle(
                normalize(Vector(2, 7, 9)),
                toRad(30)
                );
    Vector v1 = v0;
    for (int i = 0; i < 12; ++i) {
        v1 = rotate(q, v1);
    }
    assert(isEqual(v0, v1));

    Quaternion q1 = Quaternion::identity();
    for (int i = 0; i < 12; ++i) {
        q1 = q1*q;
    }
    assert(isEquivalent(q1, Quaternion::identity()));
}

void unitTestQuaternionMix() {
    Versor a = normalize(Vector(2, 7, 9));
    Quaternion q1 = Quaternion::axisAngle(a, toRad(30));

    Quaternion q2 = Quaternion::axisAngle(a, toRad(90));

    assert(isEquivalent(
               mix(q1, q2, 0.5f),
               Quaternion::axisAngle(a, toRad(60)
           )));

    assert(isEquivalent(
               mix(Quaternion::identity(), q2, 0.5f),
               Quaternion::axisAngle(a, toRad(45)
           )));

    assert(isEquivalent(
               mix(q1*q1*q1, q1, 0.5f),
               q1*q1
           ));
}

int main() {
    Vector v, w(2, 3, 5);
    v[0] = 2;
    //Scalar foo = v[0]+5;
    v = v + w;
    unitTestSumIsCommutative();
    unitTestCross();
    //std::cout << rayReflection(v , w).x << std::endl;
    unitTestVersor();
    unitTestAngle();
    unitTestQuaternionRotation();
    unitTestQuaternionRotation2();
    unitTestQuaternionMix();

    return 0;
}
